local map={
	"1111111111111111",
	"1   12         1",
	"1 111111111111 1",
	"1 1      1   1 1",
	"1 1 1111 1 1 1 1",
	"1 1   1  1 1   1",
	"1 111 111111 111",
	"1   1          1",
	"111 11111111 1 1",
	"1 1          1 1",
	"1 111111111111 1",
	"1   1 1    1   1",
	"1 111 1 1111 111",
	"1     1    1   1",
	"1 11111111 111 1",
	"1              1",
	"1111114  4111111",
	"1              1",
	"1              1",
	"1  3  3  3  3  1",
	"1              1",
	"1              1",
	"1  2  5  5  2  1",
	"1              1",
	"1              1",
	"1  2  5  5  2  1",
	"1              1",
	"1              1",
	"1  3  2  2  3  1",
	"1              1",
	"1              1",
	"1111111111111111",
}

---------------------------------------------------------------------
	local s = {
		grabbed = false,
		mapW = #map[1],
		mapH = #map,
		colorSchemes = {
			{0,8}, -- White+gray
			{3,11}, -- Blue
			{6,14}, -- Red
			{5,13}, -- Green
			{4,1}, -- Yellow/orange
		},
		
		-- Player X and Y
		px = 9,
		py = 25,
		
		dir=0,
		fx, fy,
		
		-- Player speed
		speed=0.1,
		
		-- Keyboard turning speed
		turnSpeed=4,
		
		-- Turning speed for mouse
		mouseSpeed=0.5,
		
		-- Borrowed from CCLites
		COLOUR_CODE = {
			[1] = {240, 240, 240},
			[2] = {242, 178, 51},
			[4] = {229, 127, 216},
			[8] = {153, 178, 242},
			[16] = {222, 222, 108},
			[32] = {127, 204, 25},
			[64] = {242, 178, 204},
			[128] = {76, 76, 76},
			[256] = {153, 153, 153},
			[512] = {76, 153, 178},
			[1024] = {178, 102, 229},
			[2048] = {37, 49, 146},
			[4096] = {127, 102, 76},
			[8192] = {87, 166, 78},
			[16384] = {204, 76, 76},
			[32768] = {0, 0, 0},
		},
		
		-- If true, redraw
		dirty = true,
		
		-- FOV
		FOV = 0.55,
	}

	-- Screen Dimensions
	s.w, s.h = love.window.getDimensions()
	
	s.centerX, s.centerY = math.floor((s.w+1)/2), math.floor((s.h+1)/2)

	-- FOV
	s.screenDist = s.FOV * s.w
	
	-- Middle of screen
	s.prevX, s.prevY = s.w/2, s.h/2
		
---------------------------------------------------------------------


-- Changes
local realMap = {}
love.mouse.setVisible(false)
local cos, sin, floor = math.cos, math.sin, math.floor
local scan = {}

for y=1,s.mapH do
	local rc = {}
	local r = map[y]
	
	for x=1,s.mapW do
		rc[x] = tonumber(r:sub(x,x))
	end
	
	realMap[y] = rc
end

local function cast(cx,cy,angle)
	-- Direction vector
	local vx,vy=cos(angle), sin(angle)
	local slope=vy/vx
	
	-- Next distance, x and y axis points
	local ndx, ndy
	
	-- Steps, distance and block
	local dsx, dsy, bsx, bsy
	
	if vx < 0 then
		local x = ( cx % 1 )
		bsx = -1
		ndx = ( x * x * ( 1 + slope * slope ) ) ^ 0.5
		dsx = ( ( 1 + slope * slope ) ) ^ 0.5
	else
		local x=1-(cx%1)
		bsx=1
		ndx=(x*x*(1+slope*slope))^0.5
		dsx=((1+slope*slope))^0.5
	end

	if vy<0 then
		local y=(cy%1)
		bsy=-1
		ndy=(y*y*(1+1/(slope*slope)))^0.5
		dsy=((1+1/(slope*slope)))^0.5
	else
		local y=1-(cy%1)
		bsy=1
		ndy=(y*y*(1+1/(slope*slope)))^0.5
		dsy=((1+1/(slope*slope)))^0.5
	end

	local x,y=floor(cx),floor(cy)
	local cury = realMap[y]
	while x>0 and x<=s.mapW and y>0 and y<=s.mapH do
		local hitD
		local isX
		if ndx<ndy then
			-- X crossing is next
			x=x+bsx
			isX=true
			hitD=ndx
			ndx=ndx+dsx
		else
			y=y+bsy
			cury = realMap[y]
			isX=false
			hitD=ndy
			ndy=ndy+dsy
		end
		local wall=cury[x]
		if wall then
			return s.colorSchemes[wall][isX and 1 or 2], hitD
		end
	end  
end

local function turn(amt)
	s.dir=s.dir+amt
	s.fx,s.fy=math.cos(math.rad(s.dir)), math.sin(math.rad(s.dir))
end

turn(0)

-- Build table of angles and base distances per scanline
local function scanMap()
	scan = {}
	for x = 1, s.w do
		local t={}
		scan[x] = t
		
		t.angle = math.atan2(x - s.centerX, s.screenDist)
		t.dist = ((x - s.centerX)^2 + s.screenDist^2)^.5 / s.screenDist
	end
end
scanMap()

local function redraw()
	for x = 1, s.w do
		local wall, dist = cast(s.px, s.py, math.rad(s.dir) + scan[x].angle)
		if wall then
			-- Calc wall height based on distance
			local height = scan[x].dist / dist
			height = math.floor(math.min(height * s.centerY, (s.h + 1) / 2))
			love.graphics.setColor(s.COLOUR_CODE[2 ^ wall])
			love.graphics.line(x - 0.5, s.centerY - height - 0.5, x - 0.5, s.centerY + height + 0.5)
		end
	end
	
	love.graphics.setColor(255, 255, 255)
end

local function clampCollision(x,y,radius)
	-- Am I *in* a block?
	local gx,gy=math.floor(x),math.floor(y)
	if map[gy]:sub(gx,gx)~=" " then
		-- I am. Complete fail, do nothing.
		return x,y
	end
	
	-- ok, check the neighbors.
	local right=math.floor(x+radius)>gx
	local left=math.floor(x-radius)<gx
	local front=math.floor(y-radius)<gy
	local back=math.floor(y+radius)>gy
	
	local pushed=false
	
	if right and map[gy]:sub(gx+1,gx+1)~=" " then
		-- push left
		pushed=true
		x=gx+1-radius
	elseif left  and map[gy]:sub(gx-1,gx-1)~=" " then
		-- push right
		pushed=true
		x=gx+radius
	end
	
	if front and map[gy-1]:sub(gx,gx)~=" " then
		-- push back
		pushed=true
		y=gy+radius
	elseif back and map[gy+1]:sub(gx,gx)~=" " then
		-- push forward
		pushed=true
		y=gy+1-radius
	end
 
	-- if I wasn't pushed out on any side, I might be hitting a corner
	if not pushed then
		-- square rad
		local r2=radius^2
		local pushx,pushy=0,0
		if left then
			if front and map[gy-1]:sub(gx-1,gx-1)~=" " then
				-- check front-left
				local dist2=(gx-x)^2+(gy-y)^2
				if dist2<r2 then
					local pushd=(r2-dist2)/2^.5
					pushx,pushy=pushd,pushd
				end
			elseif back and map[gy+1]:sub(gx-1,gx-1)~=" " then
				local dist2=(gx-x)^2+(gy+1-y)^2
				if dist2<r2 then
					local pushd=(r2-dist2)/2^.5
					pushx,pushy=pushd,-pushd
				end
			end
		elseif right then
			if front and map[gy-1]:sub(gx+1,gx+1)~=" " then
				-- check front-left
				local dist2=(gx+1-x)^2+(gy-y)^2
				if dist2<r2 then
					local pushd=(r2-dist2)/2^.5
					pushx,pushy=-pushd,pushd
				end
			elseif back and map[gy+1]:sub(gx+1,gx+1)~=" " then
				local dist2=(gx+1-x)^2+(gy+1-y)^2
				if dist2<r2 then
					local pushd=(r2-dist2)/2^.5
					pushx,pushy=-pushd,-pushd
				end
			end
		end
		x=x+pushx
		y=y+pushy
	end
	
	return x,y
end

function love.update(dt)
	local modifier = math.min(dt / 0.05,1)
	
	if love.keyboard.isDown("left") then
		turn(-s.turnSpeed * modifier)
		s.dirty = true
	end
	
	if love.keyboard.isDown("right") then
		turn(s.turnSpeed * modifier)
		s.dirty = true
	end
	
	if love.keyboard.isDown("up") or love.keyboard.isDown("w") then
		s.px = s.px + s.fx * s.speed * modifier
		s.py = s.py + s.fy * s.speed * modifier
		s.dirty = true
	end
	
	if love.keyboard.isDown("down") or love.keyboard.isDown("s") then
		s.px = s.px - s.fx * s.speed * modifier
		s.py = s.py - s.fy * s.speed * modifier
		s.dirty = true
	end
	
	if love.keyboard.isDown("a") then
		s.px = s.px + s.fy * s.speed * modifier
		s.py = s.py - s.fx * s.speed * modifier
		s.dirty = true
	end
	
	if love.keyboard.isDown("d") then
		s.px = s.px - s.fy * s.speed * modifier
		s.py = s.py + s.fx * s.speed * modifier
		s.dirty = true
	end
	
	if love.keyboard.isDown("escape") then
		love.event.quit()
	end
	
	if not s.grabbed then
		local mX, mY = love.mouse.getPosition()
		if mX ~= s.prevX or mY ~= s.prevY then
			if mX < s.prevX then
				turn(-(s.mouseSpeed * (s.prevX - mX)) * modifier)
				s.dirty = true
			end
			
			if mX > s.prevX then
				turn((s.mouseSpeed * (mX - s.prevX)) * modifier)
				s.dirty = true
			end
		end
		
		love.mouse.setPosition( s.w/2, s.h/2 )
	end
end

love.graphics.setLineStyle("rough")
love.graphics.setBackgroundColor(76, 76, 76)

local present = love.graphics.present
local clear = love.graphics.clear
love.graphics.clear = function() end
love.graphics.present = function() end

function love.draw()
	s.px, s.py = clampCollision(s.px, s.py, .25)
	if s.dirty then
		clear()
		redraw()
		present()
		s.dirty = false
	end
end

function love.keypressed(key, uni)
	if key == "q" then
		if s.grabbed == true then s.grabbed = false elseif s.grabbed == false then s.grabbed = true end
		
		love.mouse.setVisible(s.grabbed)
	end
end

function love.mousepressed(x, y, key)
	if s.grabbed then
		if key == "l" then s.FOV = s.FOV + 0.5 end
		if key == "r" then s.FOV = s.FOV - 0.5 end
		
		s.screenDist = s.FOV * s.w
		scanMap()
		s.dirty = true
	end
end